//! The Rear error wrapper

use std::error::Error;
use std::fmt::{Display, Formatter, Result as FmtResult};
use std::result::Result as StdResult;

/// A `std::result::Result` wrapper that simply has a `Rear` as the error.
pub type Result<T> = StdResult<T, Rear>;

/// A `Rear` is simply a message paired with the error.
#[derive(Debug)]
pub struct Rear {
	error: Box<dyn Error>,
	message: String,
}

impl Rear {
	/// Return a reference to the error.
	pub fn error(&self) -> &Box<dyn Error> {
		&self.error
	}

	/// Return the message as a `&str`.
	pub fn message(&self) -> &str {
		&self.message
	}

	/// Create a new `Rear` from an error implementing `std::error::Error` with
	/// a message from anything that implementes `Into<String>`.
	pub fn new<E: Into<Box<dyn Error>>, S: Into<String>>(error: E, message: S) -> Self {
		let error = error.into();
		let message = message.into();

		Self { error, message }
	}
}

impl Display for Rear {
	fn fmt(&self, f: &mut Formatter<'_>) -> FmtResult {
		if !self.message.is_empty() {
			write!(f, "{}", self.error)
		} else {
			write!(f, "{}\n{}", &self.message, &self.error)
		}
	}
}

impl From<Box<dyn Error>> for Rear {
	fn from(error: Box<dyn Error>) -> Self {
		Self { message: error.to_string(), error }
	}
}

impl From<std::io::Error> for Rear {
	fn from(error: std::io::Error) -> Self {
		Self { message: error.to_string(), error: error.into() }
	}
}

impl From<String> for Rear {
	fn from(error: String) -> Self {
		Self { message: error.clone(), error: error.into() }
	}
}

impl From<&str> for Rear {
	fn from(error: &str) -> Self {
		Self { message: error.to_string(), error: error.into() }
	}
}

#[cfg(feature = "askama")]
impl From<askama::Error> for Rear {
	fn from(error: askama::Error) -> Self {
		Self { message: error.to_string(), error: error.into() }
	}
}

#[cfg(feature = "log")]
impl From<log::SetLoggerError> for Rear {
	fn from(error: log::SetLoggerError) -> Self {
		Self { message: error.to_string(), error: error.into() }
	}
}

#[cfg(feature = "serde_json")]
impl From<serde_json::Error> for Rear {
	fn from(error: serde_json::Error) -> Self {
		Self { message: error.to_string(), error: error.into() }
	}
}

#[cfg(feature = "tokio_postgres")]
impl From<tokio_postgres::Error> for Rear {
	fn from(error: tokio_postgres::Error) -> Self {
		Self { message: error.to_string(), error: error.into() }
	}
}

#[cfg(feature = "ureq")]
impl From<ureq::Error> for Rear {
	fn from(error: ureq::Error) -> Self {
		Self { message: error.to_string(), error: error.into() }
	}
}
